import sys
sys.path.append('./ai-microcore')

import colorama
import typer
import microcore as mc

colorama.init(autoreset=True)
app = typer.Typer()
HOME_DIR = 'data'
mc.configure(
    STORAGE_PATH=HOME_DIR,
    USE_LOGGING=False,
    DOT_ENV_FILE='.env',
    LLM_DEFAULT_ARGS={
        'temperature': 0.2,
    },
)


import json
import time
from pathlib import Path

import rich
from colorama import Fore as C
from openai import APIError

import microcore as mc
from microcore.utils import file_link
from aico.core import env
from aico.diff import apply_patch_cli, create_uni_diff, apply_patch


def decide_on_change(change: dict):
    return (
        'estimation' not in change
        or 'decision' not in change['estimation']
        or "reject" not in str(change['estimation']['decision']).lower()
    )


def apply_changes(structured_result: str | dict):
    if isinstance(structured_result, str):
        structured_result = mc.storage.read_json(structured_result)
    for item in structured_result['changes']:
        try:
            if not decide_on_change(item):
                print(f"Skipping patch {item['number']}")
                continue
            print(f"Applying patch {item['number']} from {mc.storage.file_link(item['diff_file'])}")
            apply_patch(
                f"{mc.config().STORAGE_PATH}/{item['diff_file']}",
                env().project.src_path
            )
        except Exception as e:
            mc.ui.error(f"Error applying patch {item['number']}: {e}")


def apply_changes_interactive(structured_result: str | dict):
    if isinstance(structured_result, str):
        structured_result = mc.storage.read_json(structured_result)
    items = structured_result['changes']
    messages = ["Do nothing"] + [item['message'] for item in items] + ["Apply all"]
    if items:
        message = ""
        while message != "Do nothing":
            message = mc.ui.ask_choose("Choose fix to apply", messages)
            if message == "Apply all":
                for item in items:
                    print(f"Applying patch {item['diff_file']}")
                    apply_patch_cli(
                        f"{mc.config().STORAGE_PATH}/{item['diff_file']}",
                        env().project.src_path
                    )
                    return
            for item in items:
                if item['message'] == message:
                    print(f"Applying patch {item['diff_file']}")
                    apply_patch_cli(
                        f"{mc.config().STORAGE_PATH}/{item['diff_file']}",
                        env().project.src_path
                    )
                    messages.remove(message)
                    if len(messages) == 2:
                        return


def process_file(file_name: str | Path, verbose=True):
    proj = env().project
    prompt = mc.SysMsg(mc.tpl('file-review-full-content.j2', file_name=file_name, project=proj))

    class Token:
        ITEM_BEGIN = '[[---ITEM_BEGIN---]]'
        ITEM_END = '[[---ITEM_END---]]'
        NO_CHANGES = '[[---NO_CHANGES---]]'

    # File paths
    out_dir = Path(proj.work_folder) / 'out'
    base_fn = file_name.replace('/', '_').replace('.', '_')
    ai_text_fn = f"{out_dir}/{base_fn}.ai-text"
    structured_result_fn = f"{out_dir}/{base_fn}.json"

    mc.storage.clean(out_dir)

    try_num = 0
    out = ""
    while True:
        try:
            try_num += 1
            if try_num > 3:
                return {"result": "LLM_INCORRECT_RESPONSE"}
            out = mc.llm(prompt)
            mc.storage.write(ai_text_fn, out, rewrite_existing=True)
            items = []
            if not Token.ITEM_BEGIN in out:
                break
            out = out.replace(Token.NO_CHANGES, '')
            out_parts = (
                out
                .replace('\n' + Token.ITEM_END, '')
                .replace(Token.ITEM_END, '')
                .split(Token.ITEM_BEGIN + '\n')[1:]
            )
            item_num = 0
            for part in out_parts:
                item_num += 1
                updated_fn = f"{out_dir}/{base_fn}-{item_num}.orig"
                diff_fn = f"{out_dir}/{base_fn}-{item_num}.patch"
                item = dict(
                    number=item_num,
                    file_name=file_name,
                    **mc.parse(part, required_fields=['message', 'updated_file_content'])
                )
                mc.storage.write(updated_fn, item['updated_file_content'], rewrite_existing=True)
                create_uni_diff(
                    proj.src_path / file_name,
                    mc.storage.path / updated_fn,
                    mc.storage.path / diff_fn,
                    proj.src_path
                )
                item['diff_file'] = diff_fn
                item['diff'] = mc.storage.read(diff_fn)
                items.append(item)
            break
        except mc.BadAIAnswer as e:
            mc.ui.error(f"Bad AI Answer: {e}, Retrying [{try_num}]...")
            print(f"LLM Response:\n{C.GREEN}{out}")
            ...

    verbose and rich.print(items)
    structured_result = dict(
        changes=items,
        model=mc.config().MODEL,
        affected_files=[file_name]
    )
    mc.storage.write_json(structured_result_fn, structured_result, rewrite_existing=True)
    return structured_result


def file_cr_generate_changes(file_name: str | Path, verbose=True, clean_dir=True) -> str:
    prj = env().project
    print(f"Generating Code Review Changes for {prj.file_link(file_name)}...")
    prompt = mc.SysMsg(mc.tpl('cr-file-gen-changes.j2', file_name=file_name, project=prj))

    class Token:
        ITEM_BEGIN = '[[---ITEM_BEGIN---]]'
        ITEM_END = '[[---ITEM_END---]]'
        NO_CHANGES = '[[---NO_CHANGES---]]'

    # File paths
    out_dir = Path(prj.work_folder) / 'out'
    base_fn = file_name.replace('/', '_').replace('.', '_')
    ai_text_fn = f"{out_dir}/{base_fn}.ai-text"
    structured_result_fn = f"{out_dir}/{base_fn}.json"
    if clean_dir:
        mc.storage.clean(out_dir)

    try_num = 0
    out = ""
    while True:
        try:
            try_num += 1
            if try_num > 1+env().context.code_review.retries_if_nothing_found:
                return {"result": "LLM_INCORRECT_RESPONSE"}
            out = mc.llm(prompt)
            mc.storage.write(ai_text_fn, out, rewrite_existing=True)
            items = []
            if Token.ITEM_BEGIN not in out:
                if Token.NO_CHANGES in out:
                    if try_num <= env().context.code_review.retries_if_nothing_found:
                        print(
                            f"Retrying (no changes) "
                            f"[{try_num}/{env().context.code_review.retries_if_nothing_found}]..."
                        )
                        continue
                break
            out = out.replace(Token.NO_CHANGES, '')
            out_parts = (
                out
                .replace('\n' + Token.ITEM_END, '')
                .replace(Token.ITEM_END, '')
                .split(Token.ITEM_BEGIN + '\n')[1:]
            )
            item_num = 0
            for part in out_parts:
                item_num += 1
                item = dict(
                    number=item_num,
                    file_name=file_name,
                    **mc.parse(part, required_fields=['short_info', 'implementation_details'])
                )
                items.append(item)
            break
        except mc.BadAIAnswer as e:
            mc.ui.error(f"Bad AI Answer: {e.message}, Retrying [{try_num}]...")
            print(f"LLM Response:\n{C.GREEN}{out}")

    verbose and rich.print(items)
    structured_result = dict(
        changes=items,
        model=mc.config().MODEL,
        affected_files=[file_name]
    )
    mc.storage.write_json(structured_result_fn, structured_result, rewrite_existing=True)
    print("Results stored in "+mc.storage.file_link(structured_result_fn))
    return structured_result_fn


def provide_implementation(change: dict):
    proj = env().project
    file_name = change['file_name']
    item_num = change['number']
    prompt = mc.tpl(
        'task-update-file.j2',
        file_name=file_name,
        project=proj,
        task=f"Short info:\n{change['short_info']}\nImplementation details:\n{change['implementation_details']}\n"
    )
    try_num = 0
    out = ""
    while True:
        try:
            try_num += 1
            if try_num > 3:
                return {"result": "LLM_INCORRECT_RESPONSE"}
            out = mc.llm(prompt)
            fc = mc.parse(out, required_fields=['begin_new_file_content'])['begin_new_file_content']
            break
        except mc.BadAIAnswer as e:
            mc.ui.error(f"Bad AI Answer: {e.message}, Retrying [{try_num}]...")
            print(f"LLM Response:\n{C.GREEN}{out}")
        except APIError as e:
            sleep = 30
            mc.ui.error(f"API Error {e}, Retrying after {sleep} sec[{try_num}]...")
            time.sleep(sleep)
    out_dir = Path(proj.work_folder) / 'out'
    base_fn = file_name.replace('/', '_').replace('.', '_')

    updated_fn = f"{out_dir}/{base_fn}-ch{item_num}.new"
    mc.storage.write(updated_fn, fc, rewrite_existing=True)
    diff_fn = f"{out_dir}/{base_fn}-ch{item_num}.patch"

    create_uni_diff(
        proj.src_path / file_name,
        mc.storage.path / updated_fn,
        mc.storage.path / diff_fn,
        proj.src_path
    )
    change['diff_file'] = diff_fn
    change['diff'] = mc.storage.read(diff_fn)
    return change


def provide_implementations(structured_result_fn):
    structured_result = mc.storage.read_json(structured_result_fn)
    items = structured_result['changes']
    for i in range(len(items)):
        if not decide_on_change(items[i]):
            print(f"Skipping implementing rejected change #{items[i]['number']}")
            continue
        print(
            f"{C.MAGENTA} Implementing change "
            f"#{items[i]['number']}/{len(items)} [{C.WHITE} {items[i]['short_info']} {C.MAGENTA}]..."
        )
        items[i] = provide_implementation(items[i])
        mc.storage.write_json(structured_result_fn, structured_result, backup_existing=False)
    print("Implementations ready")
    return structured_result


def self_estimate_changes(structured_result_fn, rewrite = True):
    structured_result = mc.storage.read_json(structured_result_fn)
    try_num = 0
    out = ""
    old_changes = structured_result['changes']
    if len(old_changes) == 0:
        return structured_result
    while True:
        try:
            try_num += 1
            if try_num > 3:
                return {"result": "LLM_INCORRECT_RESPONSE"}

            updated_changes = mc.tpl(
                'self-estimate-changes.j2',
                data=json.dumps(structured_result, indent=2),
                project=env().project,
                file_name=structured_result['affected_files'][0]
            ).to_llm().parse_json(required_fields=['changes'])['changes']
            for i in range(len(old_changes)):
                if old_changes[i]['number'] != updated_changes[i]['number']:
                    raise Exception("Change Numbers mismatch")
                old_changes[i]['estimation'] = updated_changes[i]['estimation']
            break
        except Exception as e:
            mc.ui.error(f"Bad AI Answer: {str(e)}, Retrying [{try_num}]...")
            print(f"LLM Response:\n{C.GREEN}{out}")
    if rewrite:
        mc.storage.write_json(structured_result_fn, structured_result, backup_existing=False)
    return structured_result





import microcore as mc
from aico.core import project


def get_last_backup_number() -> int | None:
    backup_path = project().work_path / "backups"
    if not backup_path.is_dir():
        return None
    numeric_dirs = sorted(
        (int(folder.name) for folder in backup_path.iterdir() if folder.is_dir() and folder.name.isdigit()),
        reverse=True
    )
    return numeric_dirs[0] if numeric_dirs else None


def get_next_backup_folder():
    return project().work_folder / "backups" / str((get_last_backup_number() or 0) + 1)


def get_last_backup_folder():
    n = get_last_backup_number()
    return (project().work_folder / "backups" / str(n)) if n else None


def backup_src_folder():
    bckp_folder = get_next_backup_folder()
    print(f"Backing up src folder into {mc.storage.file_link(bckp_folder)}")
    ignore = project().ignore + [f"{i}*" for i in project().ignore]
    mc.storage.copy(
        project().src_folder,
        bckp_folder, ignore
    )


def rollback_src_folder():
    if not (bckp_folder := get_last_backup_folder()):
        raise ValueError("No backups found")
    print(f"Backing up src folder into {mc.storage.file_link(bckp_folder)}")
    ignore = project().ignore + [f"{i}*" for i in project().ignore]
    ignore = project().ignore + [f"{i}*" for i in project().ignore]
    mc.storage.copy(
        project().src_folder,
        bckp_folder, ignore
    )


import logging
import microcore as mc
from dataclasses import dataclass, field
from pathlib import Path
from colorama import Fore as C

from .const import WORK_FOLDER, DEFAULT_IGNORE
from .utils import list_dir


@dataclass
class CodeReviewSettings:
    retries_if_nothing_found: int = 2


@dataclass
class GenerationSettings:
    backup_src_folder: bool = False


@dataclass
class Context:
    project_root: Path
    win_patch_command: str = "\"C:\\Program Files\\Git\\bin\\bash.exe\" -c \"patch {}\""
    code_review: CodeReviewSettings = field(default=None)
    generation: GenerationSettings = field(default=None)

    @staticmethod
    def load():
        params = mc.storage.read_json('context.json')
        params['project_root'] = Path(params['project_root'])
        params['code_review'] = CodeReviewSettings(**params.get('code_review', {}))
        params['generation'] = GenerationSettings(**params.get('generation', {}))
        return Context(**params)


@dataclass
class Project:
    src_folder: str | Path
    work_folder: str | Path
    ignore: list[str]

    _files_content: dict[str, str] | None = None
    _meta = None

    @property
    def src_path(self) -> Path:
        return mc.storage.path / self.src_folder

    @property
    def work_path(self) -> Path:
        return mc.storage.path / self.work_folder

    @property
    def work_file(self) -> str | Path:
        return self.work_folder / "work.json"

    @property
    def work_data(self) -> dict:
        return mc.storage.read_json(self.work_file, default={"steps": []})

    def save_work_data(self, value: dict):
        mc.storage.write_json(self.work_file, value, backup_existing=False)

    def file_link(self, file_name: str):
        return mc.utils.file_link(self.src_path / file_name)

    @property
    def meta(self) -> dict:
        if not self._meta:
            self._meta = mc.storage.read_json(f'{self.work_folder}/meta.json')
        return self._meta

    @property
    def files(self) -> list[str]:
        return list_dir(str(self.src_path), self.ignore)

    @property
    def not_empty_files(self):
        res = []
        for f,c in self.files_content.items():
            if len(c.strip()) <= 1:
                continue
            res.append(f)
        return res

    @property
    def files_content(self) -> dict[str, str]:
        if not self._files_content:
            self._files_content = {}
            for f in self.files:
                try:
                    self._files_content[f] = mc.storage.read(f"{self.src_folder}/{f}")
                except Exception:
                    logging.error(f"{C.RED}Error reading file {f}{C.RESET}")
        return self._files_content

    def __post_init__(self):
        self.src_folder = Path(self.src_folder)
        self.work_folder = Path(self.work_folder)
    @staticmethod
    def load():
        project_root = env().context.project_root
        work_folder = project_root / WORK_FOLDER
        project_file = work_folder / "project.json"
        if mc.storage.exists(project_file):
            return Project(**mc.storage.read_json(project_file))
        else:
            params = {
                "src_folder": str(project_root),
                "work_folder": str(work_folder),
                "ignore": DEFAULT_IGNORE,
            }
            mc.storage.write_json(project_file, params)
            return Project.load()


@dataclass
class _Env:
    storage_path: str = "data"
    context: Context = field(default_factory=Context.load)
    project: Project = None
    instance: "_Env" = None

    def __post_init__(self):
        _Env.instance = self
        self.project = Project.load()


def env() -> _Env:
    return _Env.instance or _Env()  # noqa


def project() -> Project:
    return env().project


def config() -> Context:
    return env().context



WORK_FOLDER = f".{__package__}"
DEFAULT_IGNORE = [
    WORK_FOLDER,
    '.git',
    '__pycache__',
    '.idea/',
    'venv',
    '.pytest_cache',
    '.coverage',
    'coverage.xml',
    ".pylintrc",
    '.diff',
    '.patch',
]



from collections import deque
import microcore as mc
from aico.core import project
from microcore import UserMsg

mem = deque(maxlen=100)


def chat():
    p = project()
    work_file = f"{p.work_folder}/work.json"
    while q := UserMsg(input(">> ")):

        work_data = mc.storage.read_json(work_file, default={"steps": []})
        steps = work_data["steps"]


import fnmatch
import os


def list_dir(path: str, ignore=None) -> list[str]:
    if ignore is None:
        ignore = ['.git', ]
    files_list = []
    for root, dir, files in os.walk(path):
        for file in files:
            file_path = os.path.join(root, file)
            if not any(fnmatch.fnmatch(file_path, '*' + p + '*') for p in ignore):
                files_list.append(file_path.replace(str(path) + os.sep, '').replace(os.sep, '/'))
    return files_list




import os
import subprocess
from pathlib import Path


def pull(repo_path: str | Path):
    result = subprocess.run(
        ["git", "pull"],
        cwd=repo_path,
        check=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True
    )
    print(result.stdout)


def get_logs(repo_dir: str | Path, pull_changes=False, limit=20, oneline=False) -> str:
    repo_dir = Path(repo_dir)
    if not repo_dir.is_dir():
        raise ValueError(f"Directory {repo_dir} does not exist.")
    if pull_changes:
        print("Pulling the latest changes from the repository...")
        pull(repo_dir)
    print("Collecting commits...")
    result = subprocess.run(
        ["git", "log", "-n", str(limit)] + (["--oneline"] if oneline else []),
        cwd=repo_dir,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        check=True
    )

    # Return the stdout of the command
    return result.stdout



import difflib
import subprocess
import sys
from pathlib import Path
import patch
from .core import env


def create_uni_diff(original_fn, updated_fn, diff_fn, src_root):
    with open(original_fn, 'r') as file1, open(updated_fn, 'r') as file2:
        file1_lines = file1.readlines()
        file2_lines = file2.readlines()
        target_fn = (
            Path(original_fn)
            .absolute()
            .relative_to(
                Path(src_root).absolute()
            )
            .__str__()
            .replace('\\', '/')
        )
        diff = difflib.unified_diff(file1_lines, file2_lines, fromfile=f"a/{target_fn}", tofile=f"b/{target_fn}")
        with open(diff_fn, 'w') as diff_file:
            diff_file.writelines(diff)
            diff_file.write('\n')


def apply_patch_cli(patch_file_name: str | Path, root_path: str | Path = None):
    file_name = (
        Path(patch_file_name)
        .absolute()
        .relative_to(Path(root_path).absolute())
        .__str__()
        .replace('\\', '/')
    )
    root_path = str(root_path).replace('\\', '/')
    if sys.platform.startswith("win"):
        patch_command = env().context.win_patch_command
    else:
        patch_command = "patch {}"
    patch_command = patch_command.format(
        f" -p1 --binary --ignore-whitespace -F3 -f -d {root_path}/ -i {file_name}"
    )
    print(patch_command)
    subprocess.run(patch_command, shell=True)


def apply_patch_py(patch_file_name: str | Path, root_path: str | Path = None):
    p = patch.fromfile(patch_file_name)
    p.apply(root=root_path)


def apply_patch(patch_file_name: str | Path, root_path: str | Path = None):
    apply_patch_py(patch_file_name, root_path)
    # apply_patch_cli(patch_file_name, root_path)


from .bootstrap import *
from .core import project
import rich
from colorama import Fore as C

mc.use_logging()

def rollback():
    from .backup import rollback_src_folder
    rollback_src_folder()

@app.command()
def work(
        query: list[str] = typer.Argument(...),
        td: bool = typer.Option(
            False,
            help='Ask LLM to provide technical task description for user request'
        )
):
    if isinstance(query, list):
        query = " ".join(query)
    if td:
        res = make_td([query])
        query = res['technical_task_description']
        original_user_query = res['query']
    else:
        original_user_query = query

    work_data = project().work_data
    tries = 0
    while True:
        try:
            tries += 1
            out = mc.tpl('gen/work.j2', query=query, project=project(), steps=work_data['steps']).to_llm()
            parts = out.split('[[BEGIN_FILE]]\n')[1:]
            parts = [i.split('[[END_FILE]]')[0] for i in parts]
            files = [
                mc.parse(i, r"\[(.*?)\]", required_fields=['file_path', 'file_content'])
                for i in parts
            ]

            parts = out.split('[[DELETE_FILE]]\n')[1:]
            deleted_files = [i.split('\n')[0] for i in parts]
            break
        except mc.BadAIAnswer as e:
            if tries > 3:
                mc.ui.error(f"Too many tries ({tries}), exiting")
                raise e

    if files or deleted_files:
        from .backup import backup_src_folder
        backup_src_folder()

    for f in files:
        mc.storage.write(f"{project().src_folder}/{f['file_path']}", f['file_content'], backup_existing=False)

    for f in deleted_files:
        mc.storage.delete(f"{project().src_folder}/{f}")

    work_data["steps"].append({
        "query": query,
        "original_user_query": original_user_query,
        "files": [i['file_path'] for i in files]
    })
    project().save_work_data(work_data)


@app.command()
def make_td(query: list[str] = typer.Argument(...)):
    if isinstance(query, list):
        query = " ".join(query)
    work_data = project().work_data
    out = mc.tpl(
        'gen/make-td.j2',
        query=query,
        project=project(),
        steps=work_data['steps']
    ).to_llm().parse_json(required_fields=["query", "technical_task_description"])
    print(f"{C.MAGENTA}Generated Technical task description:{C.RESET}: {out['technical_task_description']}")
    return out



import threading
import time
import webbrowser
from typing import Optional

from aico.git import get_logs
from webapp.app import webapp
from .bootstrap import app
import os
import subprocess
from aico.file_processing import process_file, apply_changes_interactive, file_cr_generate_changes, \
    provide_implementations, \
    self_estimate_changes, apply_changes
import re
from aico.core import *
import microcore as mc
import typer
from colorama import Fore as C
import rich
import aico.gen
from rich.pretty import pprint


@app.command()
def use(project_name: str):
    ctx = mc.storage.read_json('context.json')
    ctx['project_root'] = project_name
    mc.storage.write_json('context.json', ctx, backup_existing=False)


@app.command()
def list_files():
    rich.print(env().project.files)


@app.command()
def describe():
    q = mc.tpl('describe.j2', project=env().project)
    out = mc.llm(q, callback=lambda text: print(text, end=''))
    mc.storage.write_json(f'{env().project.work_folder}/meta.json', out.parse_json())
    print("Metadata saved")
    # mc.tpl('collect-meta.j2', input = mc.tpl('')).to_llm(


@app.command()
def improve():
    files = env().project.files
    print(files)

    out = ""
    for f, c in env().project.files_content.items():
        out += "====== FILE: " + f + ":\n"
        out += c + "\n"

    # print(len(out.split("\n")))

    processed = mc.tpl('one-improvement.j2', input=out, temperature=0).to_llm()
    print(processed)
    parse_and_rewrite(processed)


@app.command()
def files():
    mc.configure(STORAGE_PATH='data', USE_LOGGING=False)
    files_list = env().project.not_empty_files
    outs = {}
    i = 0
    for file_name in files_list:
        i += 1
        try:
            print(f"Processing [{i}/{len(files_list)}]: {C.CYAN}{file_name}{C.RESET}...\t", end='')
            out = process_file(file_name, autoapply=True, verbose=True)
            if "file_content" in out:
                del out["file_content"]
            print(f"{C.MAGENTA}{out['result']}")
            outs[file_name] = out
        except Exception as e:
            mc.ui.error(f"Error: {e}")
            raise e
        mc.storage.write_json(f'{env().project.work_folder}/files_out.json', outs, rewrite_existing=True)
    mc.storage.write_json(f'{env().project.work_folder}/files_out.json', outs, rewrite_existing=True)


def choose_file_if_needed(file_name: str):
    if not file_name or file_name not in env().project.not_empty_files:
        if file_name:
            print(f"{C.RED} File {file_name} not exists, choose another")
        file_name = mc.ui.ask_choose("Choose file", env().project.files)
    return file_name


@app.command()
def file(file_name: str = typer.Argument(None)):
    file_name = choose_file_if_needed(file_name)
    structured_result = process_file(file_name)
    apply_changes_interactive(structured_result)


@app.command()
def files_ms(apply: bool = typer.Option(True, help="Apply changes to project files")):
    files_list = env().project.not_empty_files
    out_dir = Path(env().project.work_folder) / 'out'
    mc.storage.clean(out_dir)

    for file_name in files_list:
        file_ms(file_name, clean_dir=False, apply=apply)


@app.command()
def file_ms(file_name: str = typer.Argument(None), clean_dir=True, apply=True):
    file_name = choose_file_if_needed(file_name)
    # print(f"Generating changes for {file_name}...")
    structured_result_fn = file_cr_generate_changes(file_name, clean_dir=clean_dir)
    if not isinstance(structured_result_fn, str):
        mc.ui.error("Error generating changes: no result file")
        return
    print("Estimating changes...")
    self_estimate_changes(structured_result_fn)
    print("Working on change implementations")
    provide_implementations(structured_result_fn)
    rich.print(mc.storage.read_json(structured_result_fn))
    if apply:
        print("Applying changes...")
        apply_changes(structured_result_fn)
    print('file_ms: finished')


@app.command()
def pull():
    os.chdir(env().project.src_path)
    result = subprocess.run(["git", "pull"], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
    print(result.stdout)


def parse_and_rewrite(input_text):
    # Regular expressions for extracting sections
    info_pattern = r"=====INFO=====\s*(.*?)\s*=====FILENAME====="
    filename_pattern = r"=====FILENAME=====\s*(.*?)\s*=====CONTENT====="
    content_pattern = r"=====CONTENT=====\s*(.*?)\s*(?=====|$)"

    # Extracting sections using regex
    info_match = re.search(info_pattern, input_text, re.DOTALL)
    filename_match = re.search(filename_pattern, input_text, re.DOTALL)
    content_match = re.search(content_pattern, input_text, re.DOTALL)

    if not filename_match or not content_match:
        print("Filename or content not found in the input text.")
        return

    filename = filename_match.group(1).strip()
    content = content_match.group(1).strip()

    # Optionally, use info
    info = info_match.group(1).strip() if info_match else "No description provided."

    # Rewrite the file with the new content
    with open(env().project.src_path / filename, 'w') as file:
        file.write(content)


def open_browser():
    time.sleep(1)
    webbrowser.open("http://localhost:5000")


@app.command()
def report():
    threading.Thread(target=open_browser, daemon=True).start()
    webapp.run(debug=False, use_reloader=False)


@app.command()
def make_release_message(details: str = ""):
    logs = get_logs(env().project.src_path, limit=20)
    # language=Jinja2

    out = mc.prompt(
        """
Make a release message based on the git log.
Ignore commits before previous release.
Use Markdown syntax.
Do not include links to commits to changelog.
{{ details }}
{{ logs }}
        """,
        logs=logs,
        details=details
    ).to_llm()
    print(out)


if __name__ == "__main__":
    app()
